/*
Suppose that DS_t is the type of some data structure.

Here are six patterns for functions that use a DS_t
data structure as either an input or an output value.

The first three are used all the time.
The fourth is rarely used.
The last two are why C++ has copy constructors.

DS_t * f(...);                 // callee allocates DS_t, caller may or may not de-allocate

       f(const DS_t * x, ...); // caller allocates DS_t. DS_t is strictly an input

       f(DS_t * x, ...);       // caller allocates DS_t. DS_t may be input and/or output.

       f(DS_t ** x, ...);      // probably callee allocates DS_t and caller de-allocates it

DS_t   f(...);                 // DS_t automatically allocated in caller's local space

       f(DS_t x, ...);         // DS_t automatically allocated in callee's local space

Below are examples of each possible use of the data structure DS_t by a function.
*/
#include <stdio.h>
#include <stdlib.h>


typedef struct DS_t {
   int a;
   double b;
   char c;
} DS_t;


DS_t * f1a(int n)  /* callee allocates DS_t, caller must not de-allocate */
{
   static DS_t x;
   x.a = n;
   return &x;
}


DS_t * f1b(int n)  /* callee allocates DS_t, caller must de-allocate */
{
   DS_t * x = (DS_t *)malloc(sizeof(DS_t));
   x->a = n;
   return x;
}


void f2(const DS_t * x)  /* caller allocates DS_t, DS_t is strictly an input */
{
   printf("in f2(), x->a = %d\n", x->a);
// x->a = -1;           /* this does not compile */
   ((DS_t *)x)->a = -1; /* this does compile, and breaks the const declaration */
   DS_t * y = x;        /* the previous line is equivalent to this plus the next line */
   y->a = -1;
   x++;                 /* notice that x is NOT constant (what x points at is) */
}


void f3(DS_t * x) /* caller allocates DS_t, DS_t may be used for input and/or output */
{
   printf("in f3(), x->a = %d\n", x->a);
   x->a = 3;
}


void f4(DS_t ** x, int n)  /* callee allocates DS_t, caller must de-allocate it */
{
   *x = (DS_t *)malloc(sizeof(DS_t));
   (*x)->a = n;
}


DS_t f5(int n)  /* DS_t automatically allocated in caller's local space */
{
   DS_t x;
   x.a = n;
   return x;
}


void f6(DS_t x)  /* DS_t automatically allocated in callee's local space */
{
   printf("in f6(), x.a = %d\n", x.a);
}


int main()
{
   DS_t * ds1a_p = f1a(10);
   printf("in main(), ds1a_p->a = %d\n", ds1a_p->a);

   DS_t * ds1b_p = f1b(-10);
   printf("in main(), ds1b_p->a = %d\n", ds1b_p->a);
   free(ds1b_p);

   DS_t ds2;    /* we could dynamically allocate ds2 */
   ds2.a = 20;
   f2( &ds2 );
   printf("in main(), ds2.a = %d\n", ds2.a);

   DS_t ds3;    /* we could dynamically allocate ds3 */
   ds3.a = 30;
   f3( &ds3 );
   printf("in main(), ds3.a = %d\n", ds3.a);

   DS_t * ds4_p;
   f4( &ds4_p, 40 );
   printf("in main(), ds4_p->a = %d\n", ds4_p->a);
   free( ds4_p );

   DS_t ds5 = f5(50); /* return value from f5() is copied into ds5 */
   printf("in main(), ds5.a = %d\n", ds5.a);

   DS_t ds6;
   ds6.a = 60;
   f6(ds6); /* ds6 is copied into f6's actual parameter */

   return 0;
}
